import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Tuple

from lxml import etree

import config
from consts import TS_FORMAT_STRING
from exceptions import StopModule
from generic_module import Module


class GapModule(Module):
    def __init__(self, args):
        self.time_gap = 20.0
        self.xml_dir = config.sites_dir / config.site_name / 'ext'
        Module.__init__(self, args, mandatory_keys=tuple(), optional_keys=tuple('time_gap', ))
        self.type = 'gap'

    def add_gap(self, file_name: str):
        """ Reads XML, sorts vehicle elements by their wim/ts subkey, and writes sorted XML.
        :param file_name: XML file name.
        """
        # Parse the XML with whitespace removal for clean output
        # parser = etree.XMLParser(remove_blank_text=True)
        if not Path(file_name).is_file():
            raise StopModule(f'File "{file_name}" does not exist.')
        tree = etree.parse(file_name)
        root = tree.getroot()

        # Find all vehicle elements
        vehicles = root.xpath('/swd/site/vehicles/vehicle')
        # Sort vehicles by wim/ts child element
        vehicles_sorted = sorted(
            vehicles,
            key=lambda x: datetime.strptime(
                x.xpath('./wim/ts/text()')[0],
                TS_FORMAT_STRING
            )
        )
        if len(vehicles) != len(vehicles_sorted):
            raise StopModule('Failed to sort vehicles!')

        # Remove all vehicles from their parents
        for vehicle in vehicles:
            parent = vehicle.getparent()
            if parent is not None:
                parent.remove(vehicle)

        gaps: Dict[str, Tuple[datetime, float]] = {}  # Holds timestamp of last axle and speed of vehicle.

        with open(file_name, 'w', encoding='utf-8') as output_file:
            # Write header.
            output_file.write(f'<?xml version="1.0" ?>\n<swd version="1"><site><name>{config.site_name}</name><vehicles>\n')
            # Write vehicles with gaps.
            for vehicle in vehicles_sorted:
                lane = vehicle.find('wim/lane').text
                ts = datetime.strptime(vehicle.find('wim/ts').text, TS_FORMAT_STRING)
                v = float(vehicle.find('wim/v').text)
                if lane in gaps:  # If it's first occurrence of the lane, just save the information for next vehicle.
                    gap_seconds = (ts - gaps[lane][0]).total_seconds()
                    if gap_seconds < self.time_gap:  # Don't save it, if vehicles are more than 20 seconds apart.
                        gap = (gaps[lane][1] + v) / 2 * gap_seconds
                        node = etree.Element('integration')
                        el = etree.Element('gap')
                        el.text = str(round(gap, 4))
                        node.append(el)
                        vehicle.append(node)
                gaps[lane] = (ts + timedelta(seconds=float(vehicle.find('wim/whlbse').text) / v), v)  # Save timestamp of last axle and speed.
                output_file.write(etree.tostring(vehicle).decode().strip() + '\n')
            # Write footer.
            output_file.write('</vehicles></site></swd>')

    def run(self) -> None:
        self.alive = True
        self.end = False
        self.logger.debug('Started.')
        try:
            # Convert yesterday's file, if needed.
            try:
                self.logger.info('Checking if yesterday\'s file needs to be updated ...')
                date = datetime.now().date() - timedelta(days=1)
                # If any gap tag already exist, the file has been handled already.
                if len(etree.parse(str(self.xml_dir / f'{date.isoformat()}.xml')).getroot().xpath('/swd/site/vehicles/vehicle/integration/gap')) != 0:
                    self.logger.info('Yesterday\'s file does not need to be updated.')
                    date = datetime.now().date()
            except OSError:
                self.logger.warning(f'Could not find yesterday\'s file.')
                date = datetime.now().date()

            try:
                while True:
                    if self.end:
                        raise StopModule('Shutdown flag detected.')
                    time.sleep(30)  # There's no rush to convert, so we give some more time to make sure file is no longer in use.
                    if date != datetime.now().date():
                        file_name = f'{date.isoformat()}.xml'
                        self.logger.info(f'Updating {file_name} ...')
                        self.add_gap(self.xml_dir / file_name)
                        self.logger.info(f'{file_name} updated.')
                        date = datetime.now().date()
            except StopModule:
                raise
            except:
                self.logger.exception('Unexpected exception occurred:')
                raise StopModule('Fatal error!')
        except StopModule as e:
            self.log_stop_module(e)
        finally:
            self.logger.debug('Stopped.')
